#include <sys/stat.h>
#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <limits.h>
#include <dirent.h>
#include "buffer.h"


char *imgext[]= { ".pnm", NULL };

static int scan_dest_dir(char *dn)
{
   int max_sfx= 1;
   DIR *dir= opendir(dn);
   if (!dir) exit(fprintf(stderr,"can not open destination directory %s: %s\n",
                  dn, strerror(errno)));
   while(1)
   {
     struct dirent *entry= readdir(dir);
     char *digit;
     if (!entry) {
        if (errno) exit(fprintf(stderr,
             "can not read destination directory %s: %s\n",
                   dn, strerror(errno)));
        break;
     }
     for(digit= entry->d_name+strlen(entry->d_name)-1;
                  digit!=entry->d_name;digit--)
         if (isdigit(digit[0])) 
           break;
     if (!isdigit(digit[0])) continue;
     while(digit>entry->d_name && isdigit(digit[-1])) digit--;
     while(*digit=='0') digit++;
     int K= atoi(digit);
     if (K>max_sfx) max_sfx= K;
   }
   closedir(dir);
   return max_sfx;
}

static char *combine(char *dir,char *file)
{
  char *name= malloc(strlen(dir)+strlen(file)+2);
  sprintf(name,"%s/%s",dir,file);
  return name;
}

static void get_sfx(char *fn,char *sfx)
{
  char *K= fn+ strlen(fn);
  while(K>fn && K[-1]!='.') K--;
  if (K==fn) *sfx= 0;
  else strcpy(sfx, K-1);
}

static int move_file(char *srcdir, char *dstdir, char *file,char *pfx, int nr)
{
  int R;
  char *sfx;
  char *sf, *df;
  
  sfx= malloc(strlen(file)+1);
  get_sfx(file, sfx);

  sf= combine(srcdir, file);

  df= malloc(strlen(dstdir)+1+strlen(pfx)+10+ strlen(sfx) + 1);

  sprintf(df, "%s/%s%04d%s", dstdir, pfx, nr, sfx);
  printf("moving %s to %s: ", sf, df);
  R= rename(sf,df);
  if (R)
     printf(" fail(%s)\n", strerror(errno));
  else printf(" ok.\n");
  free(sf);
  free(df);
  return R;
}

static int file_matches_ext_set(char *fn,char **ext)
{
  int i;
  int lf,le;
  if (!ext) return 1;
  lf= strlen(fn);
  for(i=0;ext[i];i++)
  {
    if (!strcmp(ext[i],".*")) return 1;
    le= strlen(ext[i]);
    if (lf<=le) continue;
    if (!strcasecmp(fn+lf-le,ext[i])) return 1;
  }
  return 0;
}

static char **list_files(char *dn)
{
  DIR *dir;
  struct dirent *de;
  char **ptr;
  char *str;
  size_t i, ptrsize, strsize, L;
  buffer_t *buffer;

  dir= opendir(dn);
  if (!dir)
  {
    ptr= malloc(sizeof(*ptr));
    ptr[0]= NULL;
    return ptr;
  }

  buffer= buffer_new(1024);
  ptrsize= strsize= 0;
  while((de=readdir(dir)))
  {
    buffer_write(buffer, "U", 1);
    L= strlen(de->d_name)+1;
    buffer_write(buffer, de->d_name, L);
    strsize+= L+1;
    ptrsize++;
  }
  closedir(dir);

  ptr= malloc( (ptrsize+1)*sizeof(char*) + strsize );
  str= (char*) ptr + (ptrsize+1)*sizeof(char*) ;

  buffer_read(buffer, str, strsize);
  buffer_free(buffer, NULL, NULL);

  for(i=0;i<ptrsize;i++)
  {
    ptr[i]= str;
    str+= strlen(str)+1;
  }
  ptr[i]= NULL;
  
  return ptr;
}

static void file_types(char *dn, char **files)
{
  size_t i;
  size_t mlen,L, dlen;
  char *path,*name;

  mlen= 0;
  for(i=0;files[i];i++)
  {
    L= strlen(files[i]);
    if (L>mlen) mlen= L;
  }

  dlen= strlen(dn);
  path= malloc(dlen+2+mlen); // slash + null
  name= path+dlen+1;
  memcpy(path, dn, dlen);
  path[dlen]= '/';

  for(i=0;files[i];i++)
  {
    struct stat sbuf;
    strcpy(name, files[i]+1);
    if (stat(path,&sbuf)) continue;
    if (S_ISDIR(sbuf.st_mode))      files[i][0]= 'D';
    else if (S_ISREG(sbuf.st_mode)) files[i][0]= 'F';
  } 

  free(path);
}

static int filter_files_byext(int nfiles,char **files, char **suffixes)
{
  int i, last;
  char *tmp;
  if (!nfiles) return 0;
  last= nfiles-1;
  for(i=0;i<last;i++)
  {
    if (files[i][0]!='F') continue;
    if (file_matches_ext_set(files[i],suffixes)) continue;
    tmp= files[last];
    files[last]= files[i];
    files[i]= tmp;
    last--;
    i--;
  } 
  if (files[last][0]!='F' || file_matches_ext_set(files[last],suffixes))
     return last+1;
  return last;
}

static int filter_files_byhid(int nfiles,char **files,int showhid)
{
  int i, last;
  char *tmp;
  if (!nfiles) return 0;
  if (showhid) return nfiles;
  last= nfiles-1;
  for(i=0;i<last;i++)
  {
    if (files[i][1]!='.') continue;
    tmp= files[last];
    files[last]= files[i];
    files[i]= tmp;
    last--;
    i--;
  } 
  if (files[last][1]!='.') return last+1;
  return last;
}

static int comparstr(const void *a,const void* b)
{
  return strcmp(*(const char**) a, *(const char**) b);
}

static void sort_files(int nfiles, char **files)
{
  qsort(files, nfiles, sizeof(files[0]), comparstr);
}


static char* combine_path(char *dir,char *fn)
{
  char *R;
  int Ld;
  Ld= strlen(dir);
  R= malloc(Ld+strlen(fn)+2);
  if (Ld>0 && dir[Ld-1]=='/')
    sprintf(R,"%s%s",dir,fn);
  else
    sprintf(R,"%s/%s",dir,fn);
  return R;
}


static void move_files_forward(int nfiles,char **files, int start)
{
  int i;
  for(i=0;i<nfiles;i++)
  {
    if (move_file(".", "target",files[i]+1, "img", start)) return ;
    start+= 2;
  }
}

static void move_files_backward(int nfiles,char **files, int start)
{
  int i;
  for(i=0;i<nfiles;i++)
  {
    if (move_file(".", "target",files[i]+1, "img", start)) return ;
    start-= 2;
  }
}

static int numoffile(char *fn)
{
  int i;
  for(i=0;fn[i];i++)
   if (isdigit(fn[i])) break;
  while(fn[i]=='0') i++;
  if (!isdigit(fn[i])) return 0;
  return atoi(fn+i);
}

/* How it works:

  first, we scan the pages forward, starting from given number.
  when the scanner is done, it's time to scan them backwards.
  if there is a jam in the middle and the user cancels the backwards
  scanning, we simply record the fact and then exits. 

  if there is a jam in the middle of forward scanning, then the user
  is supposed to stop the forward scanning, remove papers and continue
  with the backwards scanning. OK. */

static int file_has_sfx(char *fn, char *sfx)
{
  int lf, ls;
  lf= strlen(fn);
  ls= strlen(sfx);
  if (lf<=ls) return 0;
  return !strcasecmp(fn + lf - ls, sfx);
}
   
static void get_file_list(char *dir, char ***Rfiles, int *Rnfiles)
{
  char **files;
  int nfiles;
  int i;
  files= list_files(dir);
  file_types(dir,files);
  for(nfiles=0;files[nfiles];nfiles++) ;

  for(i=0;i<nfiles;i++)
    if (files[i][0]!='F' || files[i][1]=='.' || 
         !(file_has_sfx(files[i],".pnm") || file_has_sfx(files[i],".png")
           || file_has_sfx(files[i],".jpg")))
    {
      char *t;
      t= files[i];
      files[i]=files[nfiles-1];
      files[nfiles-1]= t;
      i--;
      nfiles--;
    }

  sort_files(nfiles, files);
  *Rnfiles= nfiles;
  *Rfiles= files;
}

int main()
{
  char **files,**target;
  int nfiles, ntarget;
  int i;

  get_file_list(".", &files, &nfiles);
  if (!nfiles) exit(printf("nothing was freshly scanned.\n"));
  printf("%d input files were found.\n", nfiles);
  get_file_list("target", &target, &ntarget);

  int maxtarget;
  int num;

  maxtarget= 0;
  for(i=0;i<ntarget;i++)
  {
    num= numoffile(target[i]+1);
    if (num>maxtarget) maxtarget= num;
  }
  if (maxtarget==0)
  { // no files, so we have a fresh start. move the files forwards, starting
    // from 1.
    move_files_forward(nfiles, files, 1);
    return 0;
  }

  if (maxtarget%2==1)
  {
    // the maximum value was an odd number, so the next in sequence should
    // be the next even number. This number should be found as missing
    // so that a backward scan is selected.
    maxtarget++;
  }

  char *present;
  present= calloc(1,maxtarget+1);

  for(i=0;i<ntarget;i++)
  {
    num= numoffile(target[i]+1);
    present[num]= 1;
  }

  for(i=maxtarget;i>0;i--)
   if (!present[i]) 
   {
     move_files_backward(nfiles, files, i);
     return 0;
   }

  move_files_forward(nfiles, files, maxtarget+1);

  return 0;
}


