#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "pngio.h"

/* This program resizes an image to a smaller size, which would fit 
   on a tablet screen, for example.  */

void usage(char **argv)
{
  fprintf(stderr,"usage: %s input.png output.png target-width\n", argv[0]);
  exit(1);
}

int main(int argc,char **argv)
{
  FILE *f;
  unsigned char *inp;
  unsigned int inpw,inph;
  unsigned char *out;
  unsigned int outw,outh;
  unsigned char *data;
  unsigned int datalen;
  char *e;

  if (argc!=4) usage(argv);

  e= read_png_graymap_noalpha(argv[1], &inpw, &inph, &inp);
  if (e) exit(fprintf(stderr,"png error: %s\n",e));
  
  int tw,th;
  tw= atoi(argv[3]);
  if (tw<=100) exit(fprintf(stderr,"invalid target width: %d\n", tw));

  double R;
  R= (double) tw/inpw;

  th= R*inph;

  
  if (resize_image_g8_fixed(inp, inpw, inph, tw, th, &out, &outw, &outh))
    exit(fprintf(stderr,
        "failed to resize, somehow: input= %dx%d target=%dx%d\n", inpw,inph,
        tw,th));
  free(inp);

  e= encode_png_g8(outw,outh,out, &datalen, &data);
  if (e) exit(fprintf(stderr,"png error: %s\n",e));
  free(out);

  f= fopen(argv[2],"w");
  if (!f) exit(fprintf(stderr,"can't open %s: %s\n",argv[2], strerror(errno)));
  int P;
  P= fwrite(data, 1, datalen, f);
  if (P!=datalen)
    exit(fprintf(stderr,"write(%d) returned %d. exiting.\n", datalen,P));
  fclose(f);
  return 0;
}
