#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>

#include "gfx.h"
#include "pngio.h"
#include "resize_image.h"
#include "ptrbuf.h"
#include "firead.h"

#define PAD 5
#define BUT_H 35
#define BUT_W 80
#define BUT_Y PAD

#define VIEW_BG 0xffffff
#define TOOLBAR_BG 0xcccccc

static unsigned int scr_w, scr_h;
static int view_w, view_h, view_x, view_y;
static int port_x, port_y, pos_y, copy_w;
static int pglab, lab_x,lab_y,lab_w,lab_h;


static ipak_t *doc;
static unsigned int npa, pgnum;

static unsigned char *img; unsigned int img_w, img_h;


enum { BTN_PREV= GFXEV_LAST+1,
       BTN_NEXT, 
       BTN_UP, BTN_DOWN, BTN_FILES };

static void screen()
{
  int x;

  gfx_geometry(&scr_w, &scr_h);

  view_w= scr_w;
  view_x= 0;

  view_y= PAD + BUT_H + PAD;
  view_h= scr_h - view_y;

  gfx_fill_rect(0, 0,0, scr_w, view_y, TOOLBAR_BG);
  gfx_fill_rect(0, view_x,view_y, view_w, view_h, 0xffffff);

  x= scr_w - PAD - BUT_W;
  gfx_button(0,x, BUT_Y,BUT_W,BUT_H,"Down", BTN_DOWN);
  x-= PAD+BUT_W;
  gfx_button(0,x, BUT_Y, BUT_W, BUT_H, "Up", BTN_UP);

  x= PAD;
  gfx_button(0,x, BUT_Y, BUT_W, BUT_H, "Prev", BTN_PREV);
  x+= PAD+BUT_W;
  gfx_button(0,x, BUT_Y, BUT_W, BUT_H, "Next", BTN_NEXT);
  x+= 4*PAD+BUT_W;


  x+= BUT_W + PAD;

  lab_x= x; lab_y= BUT_Y; lab_w= 1.5*BUT_W; lab_h= BUT_H;
  pglab= gfx_label(GFX_JUST_CENTER, 2, 0x00, 0xededed, 
                   lab_x, lab_y, lab_w, lab_h, "");
  gfx_draw_rect(0, lab_x-1, lab_y-1,lab_w+2,lab_h+2, 0x888888);

  x+= lab_w + PAD;
  gfx_button(0,x, BUT_Y, BUT_W, BUT_H, "Files", BTN_FILES);

  gfx_update(0,0,scr_w,scr_h);
}


static int load_img(unsigned int P)
{
  unsigned char *big;   unsigned int big_w, big_h;
  unsigned char *small; unsigned int small_w, small_h;
  int iH;

  if (P>=npa) return 1;
  if (ipak_render_page(doc,P,&big,&big_w,&big_h)) return 1;

  if (big_w<=view_w) goto noresize;

  iH= (double) big_h / big_w  * view_w + 10;
  if (resize_image_g8_fixed(big, big_w, big_h, view_w,
                      iH, &small, &small_w, &small_h))
    goto noresize;

  if (img) free(img);
  img= small;
  img_w= small_w;
  img_h= small_h;
  free(big);
  pgnum= P;
  return 0;

noresize:
  if (img) free(img);
  img= big;
  img_w= big_w;
  img_h= big_h;
  pgnum= P;
  return 0;
}

static void initial_display()
{
  int copy_h;
  if (img_w<view_w) { port_x= (view_w-img_w)/2; copy_w= img_w; }
               else { port_x= 0; copy_w= view_w; }
  if (img_h<view_h) { port_y= (view_h-img_h)/2; copy_h= img_h; }
               else { port_y= 0; copy_h= view_h; }
  gfx_fill_rect(0, view_x,view_y, view_w, view_h, 0xffffff);

  gfx_put_graymap( 0, img, img_w, img_h,
                      0, 0, 
                      view_x + port_x, view_y +port_y, copy_w, copy_h);
  gfx_update(view_x,view_y, view_w, view_h);
  pos_y= 0;

  char buf[200];
  sprintf(buf," %d/%d ", pgnum+1, npa);
  gfx_label_set_text(GFX_UPDATE, pglab, buf);
}

static int scroll_down()
{
  if (pos_y+view_h>=img_h) return 1;
  pos_y+= view_h - view_h/10;
  if (pos_y+view_h>img_h) pos_y= img_h - view_h;
  
  gfx_put_graymap( GFX_UPDATE, img, img_w, img_h,
                      0, pos_y, 
                      view_x + port_x, view_y +port_y, copy_w, view_h);
  return 0;
}

static int scroll_up()
{
  if (pos_y==0) return 1;
  pos_y-= view_h - view_h/10;
  if (pos_y<0) pos_y= 0;
  
  gfx_put_graymap( GFX_UPDATE, img, img_w, img_h,
                      0, pos_y, 
                      view_x + port_x, view_y +port_y, copy_w, view_h);
  return 0;
}

static void next_page()
{
  if (load_img(pgnum+1)) return ;
  initial_display();
}

static void prev_page()
{
  if (pgnum==0) return ;
  if (load_img(pgnum-1)) return ;
  initial_display();
}

#ifdef KANDROID
#define DEFAULT_FILE "/mnt/extsd/YBUKE/default.tar"
#else
extern char *default_fn;
#define DEFAULT_FILE default_fn
#endif



static int wait_button_release()
{
  gfx_event_t event;
  while(1)
  {
    if (!gfx_event(NULL,&event)) continue;
    if (event.type==GFXEV_QUIT) return 1;
    if (event.type==GFXEV_BUTTON_RELEASE) return 0;
  }
}

static int show_goto()
{
  int number;
  gfx_event_t event;
  int x,y;
  int old_scr;
  int fh;
  int tw;
  int lab, btnevstart,D;
  int win_h, win_x,win_y,win_w;
  char *title;
  char buf[20];
  int R;

  title= "Go to Page";

  number= 0;
  R= 0;
  old_scr= gfx_push_screen();

#define CBU_W 70
#define CBU_H 70
#define CWIN_W (4*PAD+CBU_W*3)
#define CWIN_H (5*PAD+CBU_H*4)

  gfx_text_size(3,"9", NULL, &fh);

  win_h= PAD + fh + PAD + fh + CWIN_H;
  win_w= CWIN_W;

  win_x= (scr_w-win_w)/2;
  win_y= (scr_h-win_h)/2;

  gfx_draw_rect(0, win_x-2, win_y-2, win_w+4, win_h+4, 0x00);
  gfx_draw_rect(0, win_x-1, win_y-1, win_w+2, win_h+2, 0x888888);

  gfx_fill_rect(0, win_x, win_y, win_w, win_h, 0xededed);

  gfx_text_size(3, title, &tw,NULL);

  y= win_y + PAD;
  gfx_print(0,3, 0x00, 0xededed, win_x + (win_w-tw)/2, y,title,NULL,NULL);
  y+= fh + PAD;

  lab= gfx_label(GFX_JUST_RIGHT, 3, 0x00, 0xcccccc,
                                   win_x + PAD, y-3, win_w - 2*PAD,
                                         fh+6, "0");
  y+= fh + PAD;

  btnevstart= GFXEV_LAST+1;

  buf[1]= 0;
  x= PAD + win_x;
  buf[0]='7'; gfx_button(0, x,y,CBU_W,CBU_H,buf,btnevstart+7); x+= PAD+CBU_W;
  buf[0]='8'; gfx_button(0, x,y,CBU_W,CBU_H,buf,btnevstart+8); x+= PAD+CBU_W;
  buf[0]='9'; gfx_button(0, x,y,CBU_W,CBU_H,buf,btnevstart+9); x+= PAD+CBU_W;

  y+= CBU_H+PAD;
  x= PAD + win_x;
  buf[0]='4'; gfx_button(0, x,y,CBU_W,CBU_H,buf,btnevstart+4); x+= PAD+CBU_W;
  buf[0]='5'; gfx_button(0, x,y,CBU_W,CBU_H,buf,btnevstart+5); x+= PAD+CBU_W;
  buf[0]='6'; gfx_button(0, x,y,CBU_W,CBU_H,buf,btnevstart+6); x+= PAD+CBU_W;

  y+= CBU_H+PAD;
  x= PAD + win_x;
  buf[0]='3'; gfx_button(0, x,y,CBU_W,CBU_H,buf,btnevstart+3); x+= PAD+CBU_W;
  buf[0]='2'; gfx_button(0, x,y,CBU_W,CBU_H,buf,btnevstart+2); x+= PAD+CBU_W;
  buf[0]='1'; gfx_button(0, x,y,CBU_W,CBU_H,buf,btnevstart+1); x+= PAD+CBU_W;

  y+= CBU_H+PAD;
  x= PAD + win_x;
  buf[0]='0'; gfx_button(0, x,y,CBU_W,CBU_H,buf,btnevstart+0); x+= PAD+CBU_W;
  gfx_button(0, x,y,CBU_W,CBU_H,"Go",btnevstart+10); x+= PAD+CBU_W;
  gfx_button(0, x,y,CBU_W,CBU_H,"Del",btnevstart+11); x+= PAD+CBU_W;

  gfx_update(0,0,scr_w,scr_h);

  while(1)
  {
    if (!gfx_event(NULL,&event)) continue;
    switch(event.type)
    {
    case GFXEV_QUIT: R= 1; goto done;
    case GFXEV_BUTTON_PRESS:
       if (event.x<win_x || event.y<win_y ||
           event.x>=(win_x+win_w) || event.y>=(win_y+win_h)) 
       {
         R= wait_button_release();
         goto done;
       }
       break;
    default:
       D= event.type - btnevstart;
       if (D<0 || D>11) break;
       if (D==10) goto jump_page;
       if (D==11) number= number/10; 
             else number= number*10 + D;
       sprintf(buf,"%d",number);
       gfx_label_set_text(GFX_UPDATE, lab, buf);
       break; 
    }
  }

done:
  gfx_restore_screen(old_scr);
  gfx_update(0,0,scr_w,scr_h);
  return R;
jump_page:
  if (number>npa || number<1) goto done; 
  gfx_restore_screen(old_scr);
  load_img(number-1);
  initial_display();
  return 0;
}


#ifdef KANDROID
#define FILES_DIR "/mnt/extsd/YBUKE"
#else
extern char *files_dir;
#define FILES_DIR files_dir
#endif

static int strsfx(char *name, const char *sfx)
{
  unsigned int NL, SL,i;
  SL= strlen(sfx);
  NL= strlen(name);
  if (NL<=SL) return 0;
  for(i=0;i<SL;i++)
    if (name[NL-SL+i]!=sfx[i]) return 0;
  return 1;
}


static char *choose_file()
{
  DIR *dir;
  struct dirent *de;
  ptrbuf_t *filebuf;
  char **files; size_t nfiles;
  int i,j;
  TUPiis_t *entries;
  char *fn;
  int R;

  dir= opendir(FILES_DIR);
  if (!dir) return NULL;
  filebuf= ptrbuf_new(100);
  while((de=readdir(dir)))
  {
    if (strsfx(de->d_name,".tar"))
     ptrbuf_put(filebuf, strdup(de->d_name));
  }
  closedir(dir);
  ptrbuf_free(filebuf, (void***) &files, &nfiles);
  if (nfiles==0) return NULL;

  for(i=0;i<nfiles-1;i++)
   for(j=i+1;j<nfiles;j++)
     if (strcmp(files[i],files[j])>0)
     {
       char *t= files[i];
       files[i]= files[j];
       files[j]= t;
     }
  entries= malloc(sizeof(entries[0])*nfiles);
  for(i=0;i<nfiles;i++)
  {
    entries[i].iv1= i;
    entries[i].iv2= 0;
    entries[i].sv= strdup(files[i]);
    char *K= strrchr(entries[i].sv,'.');
    if (K) *K= 0;
  }
  R= gfx_fwmenu(3,nfiles, entries);
  for(i=0;i<nfiles;i++)
   free(entries[i].sv);
  free(entries);
  if (R==-1)
    fn= NULL;
  else fn= strdup(files[R]);
  for(i=0;i<nfiles;i++) free(files[i]);
  free(files);
  return fn;
}

static void load_file(char *fn)
{
  if (ipak_init(&doc, fn, &npa)) 
  {
    gfx_print(GFX_UPDATE,2, 0, 0xffffff, 30, 200, ipak_error(doc), NULL,NULL);
  } 
  if (npa==0) 
  {
    gfx_print(GFX_UPDATE,2, 0, 0xffffff, 30, 200, "No pages", NULL,NULL);
  }
  pgnum= 0;

  if (load_img(0))
  {
    img_w= 20;
    img_h= 20;
    img= calloc(1,400);
  }
  initial_display();
}

static void switch_file()
{
  char *fn;
  char buf[2000];
  fn= choose_file();
  if (!fn) return ;

  sprintf(buf,"%s/%s",FILES_DIR, fn);
  free(fn);
  fn= strdup(buf);

  if (doc) ipak_close(doc);
  doc= NULL;
  npa= pgnum= 0;
  if (img) free(img);
  img= NULL;
  img_w= img_h= 0;

  load_file(fn);
}

void android_main(void *state, size_t state_size)
{
  gfx_event_t event;
  int mx,my,dy, dragging;

  dragging= 0;
  screen();

  load_file(DEFAULT_FILE);

  while(1) {
   if (!gfx_event(NULL,&event)) continue;
   switch(event.type)
   {
   case GFXEV_QUIT: return ;
   case BTN_DOWN: if (scroll_down()) next_page(); break;
   case BTN_UP: if (scroll_up()) prev_page(); break;
   case BTN_NEXT: next_page(); break;
   case BTN_PREV: prev_page(); break;
   case BTN_FILES: switch_file(); break;
   case GFXEV_BUTTON_PRESS:
      if (event.button!=1) break;
      mx= event.x; my= event.y;
      if (mx>=view_x && my>=view_y && mx<=(view_x+view_w) && my<(view_y+view_h))
         dragging= 1; 
      else if (mx>=lab_x && my>=lab_y && mx<(lab_x+lab_w) && my<(lab_y+lab_h))
        { 
          if (wait_button_release()) return ;
          if (show_goto()) return; 
        }
      break;
   case GFXEV_BUTTON_RELEASE:
      if (event.button!=1) break;
      dragging= 0;
      break;
   case GFXEV_POINTER_MOVE: 
      if (!dragging) break;
      dy= my - event.y;
      if (dy==0) break; 
      dy+= dy/2;
      my= event.y;
      if (dy<0)
      {
       if (pos_y==0) continue;
       pos_y += dy;
       if (pos_y<0) pos_y= 0;
      }
      else
      {
       if (pos_y+view_h>=img_h) continue;
       pos_y+= dy;
       if (pos_y+view_h>img_h) pos_y= img_h-view_h;
      }
      gfx_put_graymap( GFX_UPDATE, img, img_w, img_h,
                       0, pos_y, 
                       view_x + port_x, view_y +port_y, copy_w, view_h);
      break; 
   }
 }
}
